"""Count tracking logic and bot's external interface."""

from contextlib import nullcontext as nullcontext
from dataclasses import dataclass
from functools import total_ordering
from itertools import islice
from typing import Optional
from uuid import UUID
import bisect
import datetime as dt
import importlib.metadata
import itertools
import logging

from trio import CancelScope, EndOfChannel
import trio

from strikebot.updates import Command, parse_update


__version__ = importlib.metadata.version(__package__)


@dataclass
class _Update:
	id: str
	name: str
	author: str
	number: Optional[int]
	command: Optional[Command]
	ts: dt.datetime
	count_attempt: bool
	deletable: bool
	stricken: bool

	def can_follow(self, prior: "_Update") -> bool:
		"""Determine whether this count update can follow another count update."""
		return self.number == prior.number + 1 and self.author != prior.author

	def __str__(self):
		content = str(self.number)
		if self.command:
			content += "+" + self.command.name
		return "{}({} by {})".format(self.id[4 : 8], content, self.author)


@total_ordering
@dataclass
class _BufferedUpdate:
	update: _Update
	release_at: float  # Trio time

	def __lt__(self, other):
		return self.update.ts < other.update.ts


@total_ordering
@dataclass
class _TimelineUpdate:
	update: _Update
	accepted: bool

	def __lt__(self, other):
		return self.update.ts < other.update.ts

	def rejected(self) -> bool:
		"""Whether the update should be stricken."""
		return self.update.count_attempt and not self.accepted


def _parse_rfc_4122_ts(ts: int) -> dt.datetime:
	epoch = dt.datetime(1582, 10, 15, tzinfo = dt.timezone.utc)
	return epoch + dt.timedelta(microseconds = ts / 10)


def _format_update_ref(update: _Update, thread_id: str) -> str:
	url = f"https://www.reddit.com/live/{thread_id}/updates/{update.id}"
	author_md = update.author.replace("_", "\\_")  # Reddit allows letters, digits, underscore, and hyphen
	return f"[{update.number:,} by {author_md}]({url})"


def _format_bad_strike_alert(update: _Update, thread_id: str) -> str:
	return "_incorrectly stricken:_ " + _format_update_ref(update, thread_id)


def _format_curr_count(last_valid: Optional[_Update]) -> str:
	if last_valid and last_valid.number is not None:
		count_str = f"{last_valid.number + 1:,}"
	else:
		count_str = "unknown"
	return f"_current count:_ {count_str}"


async def count_tracker_impl(
	message_rx,
	api_pool: "strikebot.reddit_api.ApiClientPool",
	reorder_buffer_time: dt.timedelta,  # duration to hold some updates for reordering
	thread_id: str,
	bot_user: str,
	enforcing: bool,
	read_only: bool,
	update_retention: dt.timedelta,
	logger: logging.Logger,
) -> None:
	from strikebot.reddit_api import CorrectionUpdateRequest, DeleteRequest, ReportUpdateRequest, StrikeRequest

	enforcing = enforcing and not read_only

	buffer_ = []
	timeline = []
	last_valid: Optional[_Update] = None
	pending_strikes: set[str] = set()  # names of updates to mark stricken on arrival
	forgotten: bool = False  # whether the retention period for an update has passed during the run

	def handle_update(update):
		nonlocal last_valid

		tu = _TimelineUpdate(update, accepted = False)

		pos = bisect.bisect(timeline, tu)
		if pos != len(timeline):
			logger.warning(f"long transpo: {update.id}")

		if pos > 0 and timeline[pos - 1].update.id == update.id:
			# The pool sent this update message multiple times.  This could be because the message reached parity and
			# its bucket was dropped just before a new connection was opened and sent its own copy, but connections also
			# seem to send duplicate messages.
			return

		pred = next(
			(timeline[i].update for i in reversed(range(pos)) if timeline[i].accepted),
			None
		)
		contender = update.command is Command.RESET or update.number is not None
		if contender and pred is None and last_valid and forgotten:
			# This is a really long transpo which we can't judge valid or invalid since we've forgotten the surrounding
			# updates. The best we can do is just ignore it.
			logger.warning(f"ignoring {update.id}: no valid prior count on record")
		else:
			timeline.insert(pos, tu)
			tu.accepted = (
				update.command is Command.RESET
				or (
					update.number is not None
					and (pred is None or pred.number is None or update.can_follow(pred))
					and not update.stricken
				)
			)
			logger.debug("{} {} -> {}".format("accepted" if tu.accepted else "rejected", pred, update))
			if tu.accepted:
				# resync subsequent updates
				newly_invalid = []
				resync_last_valid = update
				converged = False
				for scan_tu in islice(timeline, pos + 1, None):
					if scan_tu.update.command is Command.RESET:
						resync_last_valid = scan_tu.update
						if last_valid:
							converged = True
					elif scan_tu.update.number is not None:
						accept = (
							(resync_last_valid.number is None or scan_tu.update.can_follow(resync_last_valid))
							and not scan_tu.update.stricken
						)
						if accept:
							assert scan_tu.accepted
							resync_last_valid = scan_tu.update
							# resync would have no effect past this point
							if last_valid:
								converged = True
						elif scan_tu.accepted:
							newly_invalid.append(scan_tu)
							if scan_tu.update is last_valid:
								last_valid = None
						scan_tu.accepted = accept
					if converged:
						break

				if converged:
					assert last_valid.ts >= resync_last_valid.ts
				else:
					last_valid = resync_last_valid

				parts = []
				unstrikable = [tu for tu in newly_invalid if tu.update.stricken]
				if unstrikable:
					parts.append(
						"The following counts are invalid:\n\n"
						+ "\n".join(" - " + _format_update_ref(tu.update, thread_id) for tu in unstrikable)
					)

				if update.stricken:
					logger.info(f"bad strike of {update.id}")
					parts.append(_format_bad_strike_alert(update, thread_id))

				if parts:
					parts.append(_format_curr_count(last_valid))
					if not read_only:
						api_pool.enqueue_request(CorrectionUpdateRequest(thread_id, "\n\n".join(parts)))

				for invalid_tu in newly_invalid:
					if not invalid_tu.update.stricken:
						if enforcing:
							api_pool.enqueue_request(StrikeRequest(thread_id, invalid_tu.update.name, invalid_tu.update.ts))
						invalid_tu.update.stricken = True
			elif update.count_attempt and not update.stricken:
				if enforcing:
					api_pool.enqueue_request(StrikeRequest(thread_id, update.name, update.ts))
				update.stricken = True

		if not read_only and update.command is Command.REPORT:
			api_pool.enqueue_request(ReportUpdateRequest(thread_id, body = _format_curr_count(last_valid)))

	async with message_rx:
		while True:
			if buffer_:
				deadline = min(bu.release_at for bu in buffer_)
				cancel_scope = CancelScope(deadline = deadline)
			else:
				cancel_scope = nullcontext()

			msg = None
			with cancel_scope:
				try:
					msg = await message_rx.receive()
				except EndOfChannel:
					break

			if msg:
				if msg.data["type"] == "update":
					payload_data = msg.data["payload"]["data"]
					stricken = payload_data["name"] in pending_strikes
					pending_strikes.discard(payload_data["name"])
					if payload_data["author"] != bot_user:
						if last_valid and last_valid.number is not None:
							next_up = last_valid.number + 1
						else:
							next_up = None
						pu = parse_update(payload_data, next_up, bot_user)
						rfc_ts = UUID(payload_data["id"]).time
						update = _Update(
							id = payload_data["id"],
							name = payload_data["name"],
							author = payload_data["author"],
							number = pu.number,
							command = pu.command,
							ts = _parse_rfc_4122_ts(rfc_ts),
							count_attempt = pu.count_attempt,
							deletable = pu.deletable,
							stricken = stricken or payload_data["stricken"],
						)
						release_at = msg.timestamp + reorder_buffer_time.total_seconds()
						bisect.insort(buffer_, _BufferedUpdate(update, release_at))
				elif msg.data["type"] == "strike":
					slot = next(
						(
							slot for slot in itertools.chain(buffer_, reversed(timeline))
							if slot.update.name == msg.data["payload"]
						),
						None
					)
					if slot:
						if not slot.update.stricken:
							slot.update.stricken = True
							if isinstance(slot, _TimelineUpdate) and slot.accepted:
								logger.info(f"bad strike of {slot.update.id}")
								if not read_only:
									body = _format_bad_strike_alert(slot.update, thread_id)
									api_pool.enqueue_request(CorrectionUpdateRequest(thread_id, body))
					else:
						pending_strikes.add(msg.data["payload"])

			threshold = dt.datetime.now(dt.timezone.utc) - update_retention

			# pull updates from the reorder buffer and process
			new_buffer = []
			for bu in buffer_:
				process = (
					# not a count
					bu.update.number is None

					# valid count
					or (last_valid is None or last_valid.number is None or bu.update.can_follow(last_valid))
					or bu.update.command is Command.RESET

					or trio.current_time() >= bu.release_at
					or bu.update.command is Command.RESET
				)

				if process:
					if bu.update.ts < threshold:
						logger.warning(f"ignoring {bu.update}: arrived past retention window")
					else:
						logger.debug(f"processing {bu.update}")
						handle_update(bu.update)
				else:
					logger.debug(f"holding {bu.update}, checked against {last_valid})")
					new_buffer.append(bu)
			buffer_ = new_buffer
			logger.debug("last count {}".format(last_valid))

			# delete/forget old updates
			new_timeline = []
			i = 0
			for (i, tu) in enumerate(timeline):
				if i >= len(timeline) - 10 or tu.update.ts >= threshold:
					break
				elif tu.rejected() and tu.update.deletable:
					if enforcing:
						api_pool.enqueue_request(DeleteRequest(thread_id, tu.update.name, tu.update.ts))
				elif tu.update is last_valid:
					new_timeline.append(tu)

			if i:
				forgotten = True
			new_timeline.extend(islice(timeline, i, None))
			timeline = new_timeline
