from __future__ import annotations
from dataclasses import dataclass
from enum import Enum
from typing import Optional
import re

from bs4 import BeautifulSoup


Command = Enum("Command", ["RESET", "REPORT"])


@dataclass
class ParsedUpdate:
	number: Optional[int]
	command: Optional[Command]
	count_attempt: bool  # either well-formed or typo
	deletable: bool


def _parse_command(line: str, bot_user: str) -> Optional[Command]:
	if line.lower() == f"/u/{bot_user} reset".lower():
		return Command.RESET
	elif line.lower() in ["sidebar count", "current count"]:
		return Command.REPORT
	else:
		return None


def parse_update(payload_data: dict, curr_count: Optional[int], bot_user: str) -> ParsedUpdate:
	# curr_count is the next number up, one more than the last count

	NEW_LINE = object()
	SPACE = object()

	# flatten the update content to plain text
	tree = BeautifulSoup(payload_data["body_html"], "html.parser")
	worklist = tree.contents
	out = [[]]
	while worklist:
		el = worklist.pop()
		if isinstance(el, str):
			out[-1].append(el)
		elif el is SPACE:
			out[-1].append(el)
		elif el is NEW_LINE or el.name == "br":
			if out[-1]:
				out.append([])
		elif el.name in ["em", "strong", "del", "span", "sup", "code", "a", "th", "td"]:
			worklist.extend(reversed(el.contents))
		elif el.name in ["ul", "ol", "table", "thead", "tbody"]:
			worklist.extend(reversed(el.contents))
		elif el.name in ["li", "p", "div", "blockquote"] or re.match(r"h[1-6]$", el.name):
			worklist.append(NEW_LINE)
			worklist.extend(reversed(el.contents))
			worklist.append(NEW_LINE)
		elif el.name == "pre":
			worklist.append(NEW_LINE)
			worklist.extend([l] for l in reversed(el.text.splitlines()))
			worklist.append(NEW_LINE)
		elif el.name == "tr":
			worklist.append(NEW_LINE)
			for (i, cell) in enumerate(reversed(el.contents)):
				worklist.append(cell)
				if i != len(el.contents) - 1:
					worklist.append(SPACE)
			worklist.append(NEW_LINE)
		else:
			raise RuntimeError(f"can't parse tag {el.name}")

	tmp_lines = (
		"".join(" " if part is SPACE else part for part in parts).strip()
		for parts in out
	)
	pre_strip_lines = list(filter(None, tmp_lines))

	# normalize whitespace according to HTML rendering rules
	# https://developer.mozilla.org/en-US/docs/Web/API/Document_Object_Model/Whitespace#explanation
	stripped_lines = [
		re.sub(" +", " ", l.replace("\t", " ").replace("\n", " ")).strip(" ")
		for l in pre_strip_lines
	]

	return _parse_from_lines(stripped_lines, curr_count, bot_user)


def _parse_from_lines(lines: list[str], curr_count: Optional[int], bot_user: str) -> ParsedUpdate:
	command = next(
		(cmd for l in lines if (cmd := _parse_command(l, bot_user))),
		None
	)
	if lines:
		# look for groups of digits (as many as possible) separated by a uniform separator from the valid set
		first = lines[0]
		match = re.match(
			"(?P<v>v)?(?P<neg>-)?(?P<num>\\d+((?P<sep>[,. \u2009]|, )\\d+((?P=sep)\\d+)*)?)",
			first,
			re.ASCII,  # only recognize ASCII digits
		)
		if match:
			raw_digits = match["num"]
			sep = match["sep"]
			post = first[match.end() :]

			zeros = False
			while len(raw_digits) > 1 and raw_digits[0] == "0":
				zeros = True
				raw_digits = raw_digits.removeprefix("0").removeprefix(sep or "")

			parts = raw_digits.split(sep) if sep else [raw_digits]
			lone = len(lines) == 1 and (not post or post.isspace())
			typo = False
			if lone:
				all_parts_valid = (
					sep is None
					or (
						1 <= len(parts[0]) <= 3
						and all(len(p) == 3 for p in parts[1:])
					)
				)
				if match["v"] and len(parts) == 1 and len(parts[0]) <= 2:
					# failed paste of leading digits
					typo = True
				elif match["v"] and all_parts_valid:
					# v followed by count
					typo = True
				elif curr_count is not None and abs(curr_count) >= 100 and bool(match["neg"]) == (curr_count < 0):
					goal_parts = _separate(str(abs(curr_count)))
					partials = [
						goal_parts[: -1] + [goal_parts[-1][: -1]],  # missing last digit
						goal_parts[: -1] + [goal_parts[-1][: -2]],  # missing last two digits
						goal_parts[: -1] + [goal_parts[-1][: -2] + goal_parts[-1][-1]],  # missing second-last digit
					]
					if parts in partials:
						# missing any of last two digits
						typo = True
					elif any(parts == p[: -1] + [p[-1] + goal_parts[0]] + goal_parts[1 :] for p in partials):
						# double paste
						typo = True

			if match["v"] or zeros or typo or (parts == ["0"] and match["neg"]):
				number = None
				count_attempt = True
				deletable = lone
			else:
				if curr_count is not None and sep and sep.isspace():
					# Presume that the intended count consists of as many valid digit groups as necessary to match the
					# number of digits in the expected count, if possible.
					digit_count = len(str(abs(curr_count)))
					use_parts = []
					accum = 0
					for (i, part) in enumerate(parts):
						part_valid = len(part) <= 3 if i == 0 else len(part) == 3
						if part_valid and accum < digit_count:
							use_parts.append(part)
							accum += len(part)
						else:
							break

					# could still be a no-separator count with some extra digit groups on the same line
					if not use_parts:
						use_parts = [parts[0]]

					lone = lone and len(use_parts) == len(parts)
				else:
					# current count is unknown or no separator was used
					use_parts = parts

				digits = "".join(use_parts)
				number = -int(digits) if match["neg"] else int(digits)
				special = (
					curr_count is not None
					and abs(number - curr_count) <= 25
					and _is_special_number(number)
				)
				deletable = lone and not special
				if len(use_parts) == len(parts) and post and not post[0].isspace():
					count_attempt = curr_count is not None and abs(number - curr_count) <= 25
					number = None
				else:
					count_attempt = True
		else:
			# no count attempt found
			number = None
			count_attempt = False
			deletable = False
	else:
		# no lines in update
		number = None
		count_attempt = False
		deletable = True

	return ParsedUpdate(
		number = number,
		command = command,
		count_attempt = count_attempt,
		deletable = deletable,
	)


def _separate(digits: str) -> list[str]:
	mod = len(digits) % 3
	out = []
	if mod:
		out.append(digits[: mod])
	out.extend(digits[i : i + 3] for i in range(mod, len(digits), 3))
	return out


def _is_special_number(num: int) -> bool:
	num_str = str(num)
	return bool(
		num % 1000 in [0, 1, 333, 999]
		or (num > 10_000_000 and "".join(reversed(num_str)) == num_str)
		or re.match(r"(.+)\1+$", num_str)  # repeated sequence
	)
